import orjson
from typing import Callable, TypeVar

T = TypeVar("T")

class Deserializers:

    deserialization_functions = {
        int: lambda val_str: int(val_str),
        float: lambda val_str: float(val_str),
        str: lambda val_str: val_str,
        list: lambda val_str: orjson.loads(val_str),
        dict: lambda val_str: orjson.loads(val_str),
        None: lambda val_str: None
        }

    @classmethod
    def get(cls, to_type: T) -> Callable[[str], T]:
        return cls.deserialization_functions.get(to_type)
    
    @classmethod
    def __get__(cls, to_type: T) -> Callable[[str], T]:
        return cls.get(to_type)
    
    @classmethod
    def add_deserializer(cls, to_type: T, function: Callable[[str], T]) -> None:
        cls.deserialization_functions.update({to_type: function})
    
    @classmethod
    def deserialize(cls, value: str, to_type: T) -> T:

        deserializer = cls.get(to_type)
        
        return deserializer(value)





import time
import json
import orjson
import datetime

data = {
    "type": "job",
    "created_at": f'{datetime.datetime(1970, 1, 1)}',
    "status": "ok",
    "payload": [i for i in range(10)],
}

def time_dumps(data: dict, dump_method, n_loops = 1000) -> float:

    now = time.time()

    for _ in range(n_loops):
        dump_method(data)

    return time.time() - now

def time_loads(data: str | bytes, loads_method, n_loops = 1000) -> float:
    now = time.time()

    for _ in range(n_loops):
        loads_method(data)
    
    return time.time() - now
    
if __name__ == '__main__':

    dt_json = time_dumps(data, lambda x: json.dumps(x), 10000)
    dt_orjson = time_dumps(data, lambda x: orjson.dumps(x).decode(), 10000)
    
    dt_loads_json = time_loads(json.dumps(data), lambda x: json.loads(x), 10000)
    dt_loads_orjson = time_loads(json.dumps(data), lambda x: orjson.loads(x), 10000)

    print(f'Timing dumps methods.\nBuilt in Json: {dt_json} s\nOrjson: {dt_orjson} s')
    print()
    print(f'Timing loads methods.\nBuilt in Json: {dt_loads_json} s\nOrjson: {dt_loads_orjson} s')


import zmq
import time
from random import randint

from zmq_requests import (
    service_request,
    Deserializers)


class ClientTest:

    def __init__(self, host: str = 'tcp://localhost', port:int = 5555, timeout_ms: int = 5000):
        
        self.host = host
        self.port = port
        self._context = zmq.Context()
        self.socket = self._context.socket(zmq.REQ)
        self.socket.connect(f'{host}:{port}')
        self.socket.RCVTIMEO = timeout_ms

        self._check_connection_with_service()
    
    def _check_connection_with_service(self) -> None:
        try: 
            self.CheckClientConnection()
        except zmq.error.Again:
            raise Exception(f'Timeout. Client cannot establish connection to server at {self.host}:{self.port}')  

    @service_request
    def CheckClientConnection(self) -> None: ...

    @service_request
    def SayHello(self) -> None: ...
    
    @service_request
    def CheckMsg(self, msg: str) -> str: ...
    
    @service_request
    def Sum(self, a: int, b: int) -> int: ...
    
    @service_request
    def Divide(self, a: float, b: float) -> float: ...
    
    @service_request
    def Add(self, a: int, b: float) -> float: ...


    @service_request
    def GetList(self) -> list: ...

    @service_request
    def GetDictionary(self) -> dict: ...

if __name__ == '__main__':

    client = ClientTest()
    
    t0 = time.time()

    for i in range(1000):
        
        client.SayHello()
        
        msg = client.CheckMsg(f'Hello from the ClientTest!')
        print(msg)

        n1 = client.Divide(1, 1) + client.Sum(i, 1.0)
        n2 = client.Sum(randint(-10, 10), randint(-10, 10))
        
        print(n1)
        print(n2)
        
        print(client.GetList())
        print(client.GetDictionary()["x"])
        
        

    delta_time = time.time() - t0
    print(f'Delta time: {round(delta_time, 3)} secs')
        


import orjson
from functools import wraps

from .deserialization import Deserializers
from .models import (
    ServiceRequest,
    ServiceResponse,
    RequestStatus)


def service_request(function: callable) -> callable:
    
    @wraps(function)
    def wrapper(*args, **kwargs) -> dict:
        
        function(*args, **kwargs)
        
        service_args = {
            **{arg: val for arg, val in zip(function.__code__.co_varnames[1:], args[1:])},
            **kwargs}
        
        
        req_socket = args[0].socket
        req_socket.send_string(ServiceRequest(function.__name__, service_args).dumps())

        response = ServiceResponse(**orjson.loads(req_socket.recv_string()))

        if response.requestStatus != RequestStatus.SUCCESS: 
            raise Exception(f'Invalid request to service {function.__name__}. {response.serviceOutput}')
        
        return Deserializers.deserialize(response.serviceOutput, function.__annotations__['return'])
        
    return wrapper





from distutils.core import setup

with open('README.md', 'r') as readme:
  long_desc = readme.read()

setup(name='zmq_requests',
      version='0.1.0',
      description='Client requests that mimic bindings',
      long_description = long_desc,
      long_description_content_type = 'text/markdown',
      author='Lucas Harim G. C.',
      author_email='harimlgc@usp.br',
      keywords = 'zmq-requests',
      packages = ['zmq_requests'],
      install_requires = [
        'pyzmq',
        'orjson']
     )


import orjson
from dataclasses import dataclass


class RequestStatus:

    ERROR = 'ERROR'
    SUCCESS = 'SUCCESS'


@dataclass
class ServiceRequest:
    serviceName: str
    serviceArgs: dict

    def dumps(self) -> str:

        return orjson.dumps({'serviceName': self.serviceName, 'serviceArgs': self.serviceArgs}).decode("utf-8")

@dataclass
class ServiceResponse:
    requestStatus: str
    serviceOutput: str